package owg.lgm;

import owg.util.iorderset.IOrderHashSet;
import owg.util.math.Clamp;
import owg.util.math.Modulo;
import owg.util.math.Vec1;
import owg.util.math.Trig;

import static owg.lgm.GMLib.*;

import java.util.ArrayList;

import owg.lgm.in.KeyEventType;
import owg.lgm.in.MB;
import owg.lgm.in.MouseEventType;
import owg.lgm.in.VK;
import owg.lgm.room.Renderable;
import owg.lgm.room.Room;
import owg.lgm.util.Destructible;
import owg.lgm.util.IterableIterator;

public class Entity extends Renderable {
	protected static long ID_COUNT = 0x100000000L;
	public static boolean REGISTER = true;
	
	/**Creation code identifier*/
	protected String ccId;
	
	public final boolean persistent;
	protected boolean active = true;
	public String persistenceSourceRoomName;
	public int persistenceSourceInstanceIndex;

	public long id;
	public double x, y;
	public double xstart, ystart;
	public double xprevious, yprevious;
	public double hspeed = 0, vspeed = 0;
	public double image_xscale = 1, image_yscale = 1;
	protected final ObjectTransform transformCache;
	protected boolean mouseHover = false;
	
	public double image_index = 0;
	public double image_speed = 1;
	
	public double gravity = 0;
	public double gravity_direction = 0;
	public double friction = 0;
	
	public Path path = null;
	public double path_positionprevious = 0;
	public double path_position = 0;
	public double path_orientation = 0;
	public double path_scale = 1;
	public double path_speed = 0;
	public int path_endaction = path_stop;
	protected double path_ref_x;
	protected double path_ref_y;
	protected double path_hspeed = 0;
	protected double path_vspeed = 0;
	
	public boolean solid;
	
	public Sprite sprite;
	public Sprite mask;
	
	public double image_angle = 0;
	public int image_blend = 0xFFFFFF;
	public double image_alpha = 1;
	
	public int[] alarm;
	public Timeline timeline;
	public double timeline_position, timeline_speed = 1;
	
	protected final IOrderHashSet<Room> rooms = new IOrderHashSet<>();

	public Entity(double x, double y, int depth, boolean persistent, boolean visible, boolean solid, String spriteName, String maskName)
	{	
		this.x = x;
		this.y = y;
		this.xstart = x;
		this.ystart = y;
		this.xprevious = x;
		this.yprevious = y;
		this.depth = depth;
		this.id = ID_COUNT++;
		
		alarm = new int[16];
		
		
		this.persistent = persistent;
		this.visible = visible;
		this.solid = solid;
		
		if(spriteName != null)
			this.sprite = Sprite.get(spriteName);
		if(maskName != null)
			this.mask = Sprite.get(maskName);
		
		transformCache = new ObjectTransform(this);
		if(REGISTER)
			Room.CURRENT.add(this);
	}

	public void initID(int id, String ccId)
	{
		this.id = id;
		this.ccId = ccId;
	}
	public long getID()
	{
		return id;
	}
	public boolean hasCCId(String ccId)
	{
		return ccId.equals(this.ccId);
	}

	public void createEvent()
	{
		//Optional
	}
	
	public void destroy()
	{
		if(!destroyed)
			destroyEvent();
		dispose();		
	}
	
	@Override
	public void dispose()
	{
		super.dispose();
		for(Room r : rooms)
			r.disposeNotify(this);
	}
	
	public void destroyEvent()
	{
		//Optional
	}
	
	public void setPrevious() 
	{
		xprevious = x;
		yprevious = y;
		path_positionprevious = path_position;
	}
	
	public void beginStep() 
	{
		//Optional
	}
	public void doTiming()
	{
		if(timeline != null && timeline_position <= timeline.getEndTime() && timeline_speed > 0)
		{
			int t0 = ceil(timeline_position);
			timeline_position += timeline_speed;
			int t1 = ceil(timeline_position);
			for(int t = t0; t<t1; t++)
			{
				timeline.execute(t);
			}
			if(timeline_position > timeline.getEndTime()) 
				timelineEnd();
		}
		
		for(int i = 0; i<alarm.length; i++)
		{
			if(alarm[i] >= 0)
			{
				if(alarm[i] > 0)
				{
					alarm[i]--;
					if(alarm[i] == 0)
					{
						alarm[i] = -1;
						alarmTriggered(i);
					}
				}
				else
					alarm[i] = -1;
			}
		}
	}
	public void timelineEnd()
	{
		//Optional
	}
	
	public void step() {
		//Optional
	}
	public void doPhysics()
	{
		if(path != null)
		{
			hspeed = 0;
			vspeed = 0;
			
			double[] verts = path.getVerts();
			double total = verts[verts.length-1];

			double c = Trig.cos(path_orientation);
			double s = Trig.sin(path_orientation);
			
			path_position = Clamp.clamp(path_position, 0, 1);
			double absPosition = path_position * total;
			double index = Path.search(verts, absPosition);//sub-optimal
			double momentum = Path.lerp(verts, index, 2);
			
			absPosition += path_speed * momentum;
			
			boolean end = path_speed > 0 ? absPosition >= total : absPosition <= 0;
			if(end)
			{
				
				switch (path_endaction)
				{
				case path_restart:
					absPosition = Modulo.mod(absPosition, total);
					break;
				case path_continue:
					absPosition = Modulo.mod(absPosition, total);
					double px = (verts[verts.length-4] - verts[0]) * path_scale;
					double py = (verts[verts.length-3] - verts[1]) * path_scale;
					path_ref_x +=   px*c + py*s;
					path_ref_y += - px*s + py*c;
					break;
				case path_reverse:
					if(path_speed > 0)
						absPosition = Math.max(0, total*2-absPosition);
					else
						absPosition = Math.min(total, -absPosition);
					path_speed = -path_speed;
					break;
				default:
					absPosition = total;
					path = null;
					break;
				}
			}
			path_position = absPosition / total;
			assert path_position >= 0 && path_position <= 1 : this+": Bad path position: "+path_position+"/"+absPosition+"/"+total+"/"+path_endaction;
			
			index = Path.search(verts, absPosition);
			double px = (Path.lerp(verts, index, 0) - verts[0]) * path_scale;
			double py = (Path.lerp(verts, index, 1) - verts[1]) * path_scale;
			
			double oldX = x, oldY = y;
			x = path_ref_x + px*c + py*s;
			y = path_ref_y - px*s + py*c;
			path_hspeed = x-oldX;
			path_vspeed = y-oldY;
			
			if(end)
				endOfPath();
		}
		else
		{
			if(friction != 0)
			{
				double sp = getSpeed();
				if(sp < friction)
				{
					hspeed = 0;
					vspeed = 0;
				}
				else if(sp > 0)
				{
					hspeed *= 1 - friction/sp;
					vspeed *= 1 - friction/sp;
				}
			}
			if(gravity != 0)
				motion_add(gravity_direction, gravity);
			
			x += hspeed;
			y += vspeed;
		}
		
		Room room = Room.CURRENT;
		validateTransform();
		if(transformCache.x0 > room.info.width || transformCache.y0 > room.info.height ||
				transformCache.x1 < 0 || transformCache.y1 < 0) {
			outsideRoom();
			validateTransform();
		}
		if(transformCache.x1 > room.info.width || transformCache.y1 > room.info.height ||
				transformCache.x0 < 0 || transformCache.y0 < 0)
			intersectBoundary();
	}
	
	public static final int path_stop = 0;
    public static final int path_restart = 1;
    public static final int path_continue = 2;
    public static final int path_reverse = 3;
    
    public void path_start(Path path, double speed, int endaction, boolean absolute)
	{
    	this.path = path;
    	path_speed = speed;
    	path_position = 0;
    	path_endaction = endaction;	
    	if(absolute) {
    		double[] verts = path.getVerts();
    		path_ref_x = verts[0];
    		path_ref_y = verts[1];
    	} else {
    		path_ref_x = x;
    		path_ref_y = y;
    	}
	}
    public void path_end()
	{
		path = null;
		path_hspeed = 0;
		path_vspeed = 0;
	}

	public int image_number() {
		if(sprite != null)
			return sprite.number;
		else
			return 1;
	}

	public void alarmTriggered(int alarmIndex)
	{
		//Optional
	}

	public void event_user(int eventIndex)
	{
		//Optional
	}

	public void endStep() 
	{
		//Optional
	}

	public void doAnimation()
	{
		image_index += image_speed;
		if(image_index < 0 || image_index >= image_number())
		{
			animationEnd();
			image_index = Modulo.mod(image_index, 0, image_number());
		}
	}
	
	public void image_single(double index)
	{
		image_speed = 0;
		image_index = index;
	}
	
	public void animationEnd()
	{
		//Optional
	}
	
	public void endOfPath()
	{
		//Optional
	}
	
	@Override
	public void render() {
		if(sprite != null)
			draw_sprite_ext(sprite, image_index, x,y, image_xscale, image_yscale, image_angle, image_blend, image_alpha);
	}
	
	public void outsideRoom()
	{
		//Optional
	}
	public void intersectBoundary()
	{
		//Optional
	}
	
	public void roomStart()
	{
		//Optional
	}
	public void roomEnd()
	{
		//Optional
	}
	public void gameStart()
	{
		//Optional
	}
	public void gameEnd()
	{
		//Optional
	}

	
	public void checkMouseHover()
	{
		if(position_meeting(mouse_x(), mouse_y(), this))
		{
			if(!mouseHover)
			{
				mouseEntered();
				mouseHover = true;
			}
			mouseHovering();
		}
		else if(mouseHover)
		{
			mouseExited();
			mouseHover = false;
		}
	}
	
	public void mouseEntered()
	{
		//Dummy implementation	
	}
	public void mouseHovering()
	{
		//Dummy implementation	
	}
	public void mouseExited()
	{
		//Dummy implementation	
	}
	public Sprite getMask()
	{
		if(mask != null)
			return mask;
		return sprite;
	}
	
	public int sprite_width()
	{
		if(sprite == null)
			return 0;
		return sprite.width;
	}
	public int sprite_height()
	{
		if(sprite == null)
			return 0;
		return sprite.height;
	}
	public int sprite_xoffset()
	{
		if(sprite == null)
			return 0;
		return sprite.xOffset;
	}
	public int sprite_yoffset()
	{
		if(sprite == null)
			return 0;
		return sprite.yOffset;
	}
	
	public boolean place_free(double x, double y, Class<?> type, boolean solid)
	{
		if(getMask() == null)
    		return true;
		validateTransform();
		
		for(Object d : Room.CURRENT.active.getInstances(type))
		{
			if(d == this)
				continue;
			Entity e = (Entity)d;
			if(solid && !e.solid)
				continue;
			if(e.getMask() == null)
				continue;
			e.validateTransform();
			if(placeMeetingImpl(x, y, e))
				return false;
		}
		return true;
	}

    /**Skips transformUpdates and self check*/
    protected boolean placeMeetingImpl(double x, double y, Entity e) {
    	double xoff = x-this.x;
    	double yoff = y-this.y;
    	
    	return getMask().touches(transformCache, xoff, yoff, image_index, e.getMask(), e.transformCache, e.image_index);
	}
	public boolean place_free(double x, double y) {
		return place_free(x, y, Object.class, true);
    }
    public boolean place_empty(double x, double y) {
    	return place_free(x, y, Object.class, false);
	}
    
    protected static final ArrayList<Entity> meetingCache = new ArrayList<>(64);
    
    public static<T extends Entity> T instance_nearest(double x, double y, Class<T> obj) {
    	T r = null;
    	double msqd = Double.POSITIVE_INFINITY;
    	instances(meetingCache, obj);
    	for(int i = 0; i<meetingCache.size(); i++)
    	{
    		Entity e = meetingCache.get(i);
    		double sqd = Trig.sqDist(x, y, e.x, e.y);
			if(sqd <= msqd)
    		{
    			r = obj.cast(e);
    			msqd = sqd;
    		}
    	}
    	meetingCache.clear();
		return r;
    }
    
    public static<T extends Entity> T instance_furthest(double x, double y, Class<T> obj) {
    	T r = null;
    	double msqd = Double.NEGATIVE_INFINITY;
    	instances(meetingCache, obj);
    	for(int i = 0; i<meetingCache.size(); i++)
    	{
    		Entity e = meetingCache.get(i);
    		double sqd = Trig.sqDist(x, y, e.x, e.y);
			if(sqd >= msqd)
    		{
    			r = obj.cast(e);
    			msqd = sqd;
    		}
    	}
    	meetingCache.clear();
		return r;
    }
        
    public<T extends Entity> T instance_place(double x, double y, Class<T> obj) {
    	T r = null;
    	validateTransform();
    	instances(meetingCache, obj);
    	for(int i = 0; i<meetingCache.size(); i++)
    	{
    		Entity e = meetingCache.get(i);
    		if(e != this && e.getMask() != null)
    		{
	    		e.validateTransform();
	    		if(placeMeetingImpl(x, y, e))
	    		{
	    			r = obj.cast(e);
	    			break;
	    		}
    		}
    	}
    	meetingCache.clear();
		return r;
	}

    protected static Entity collisionSubject;
    protected static int collisionSolid = 0;
	protected static final ArrayList<Entity> collisionCache = new ArrayList<>(64);
    protected static int collisionIndex;
    @SuppressWarnings("rawtypes")
	protected static IterableIterator collisionIterator = new IterableIterator()
	{
		@Override
		public boolean hasNext()
		{
			collisionSubject.scanForCollision();
			return collisionIndex < collisionCache.size();
		}
		@Override
		public Object next()
		{
			Entity r = collisionCache.get(collisionIndex);
			//ugly game maker thing
			if(r.solid) {
		    	collisionSolid = 2;
		    	collisionSubject.x = collisionSubject.xprevious;
		    	collisionSubject.y = collisionSubject.yprevious;
			}
    		collisionIndex++;
			return r;
		}
	};
	protected void scanForCollision()
	{
		if(collisionSolid == 2)
		{
			x += hspeed;
			y += vspeed;
			collisionSolid = 1;
		}
		
		collisionSubject.validateTransform();
		while(collisionIndex < collisionCache.size())
    	{
    		Entity e = collisionCache.get(collisionIndex);
    		if(e != this && e.getMask() != null)
    		{
	    		e.validateTransform();
	    		if(placeMeetingImpl(x, y, e))
	    			return;
    		}
    		collisionIndex++;
    	}
	}

	@SuppressWarnings("unchecked")
	protected<T extends Entity> IterableIterator<T> getCollisions(Class<T> obj)
	{
    	collisionCache.clear();
    	instances(collisionCache, obj);
    	collisionSubject = this;
    	collisionIndex = 0;
		return collisionIterator;
	}
    
    public boolean place_meeting(double x, double y, Class<? extends Entity> obj) {
    	boolean r = false;
    	validateTransform();
    	instances(meetingCache, obj);
//    	long t0 = System.nanoTime();
    	for(int i = 0; i<meetingCache.size(); i++)
    	{
    		Entity e = meetingCache.get(i);
    		if(e != this && e.getMask() != null)
    		{
	    		e.validateTransform();
	    		if(placeMeetingImpl(x, y, e))
	    		{
	    			r = true;
	    			break;
	    		}
    		}
    	}
//    	long t1 = System.nanoTime();
//    	System.out.println(getClass()+" "+(t1-t0)/1000);
    	meetingCache.clear();
		return r;
	}
    public boolean place_meeting(double x, double y, Entity e) {
    	if(e == this || e.getMask() == null)
    		return false;
    	validateTransform();
    	e.validateTransform();
    	return placeMeetingImpl(x, y, e);
    }
	
    public static boolean position_free(double x, double y, Class<?> type, boolean solid) {
    	for(Destructible d : Room.CURRENT.active)
    	{
    		if(type.isInstance(d))
    		{
    			Entity e = (Entity) d;
    			Sprite m = e.getMask();
    			if(m != null)
    			{
	    			ObjectTransform tf = e.validateTransform();
	    			int mi = Modulo.mod((int)e.image_index, m.number);
	    			if(tf.x0 <= x && tf.x1 >= x && tf.y0 <= y && tf.x1 >= y && m.getBit(tf, mi, x, y))
	    				return false;
    			}
    		}
    	}
    	return true;
	}
    public static boolean position_free(double x, double y) {
		return position_free(x, y, Object.class, true);
	}
    public static boolean position_empty(double x, double y) {
    	return position_free(x, y, Object.class, true);
	}
    public static boolean position_meeting(double x, double y, Class<? extends Entity> obj, Entity exception) {
    	boolean r = false;
    	instances(meetingCache, obj);
    	for(int i = 0; i<meetingCache.size(); i++)
    	{
    		Entity e = meetingCache.get(i);
    		if(e != exception)
    		{
	    		e.validateTransform();
	    		if(positionMeetingImpl(x, y, e))
	    		{
	    			r = true;
	    			break;
	    		}
    		}
    	}
    	meetingCache.clear();
		return r;
	}
    public boolean position_meeting(double x, double y, Class<? extends Entity> obj) {
    	return position_meeting(x, y, obj, this);
	}

    public static<T extends Entity> T instance_position(double x, double y, Class<T> obj, Entity exception) {
    	T r = null;
    	instances(meetingCache, obj);
    	for(int i = 0; i<meetingCache.size(); i++)
    	{
    		Entity e = meetingCache.get(i);
    		if(e != exception)
    		{
	    		e.validateTransform();
	    		if(positionMeetingImpl(x, y, e))
	    		{
	    			r = obj.cast(e);
	    			break;
	    		}
    		}
    	}
    	meetingCache.clear();
		return r;
	}
    public <T extends Entity> T instance_position(double x, double y, Class<T> obj) {
    	return instance_position(x, y, obj, this);
    }
        
    
    /**Skips transformUpdates and self check*/
    protected static boolean positionMeetingImpl(double x, double y, Entity e) {
    	if(e.transformCache.lx1 <= e.transformCache.lx0)
    		return false;
    	if(e.transformCache.ly1 <= e.transformCache.ly0)
    		return false;
    	Sprite m = e.getMask();
    	int mi = Modulo.mod((int)e.image_index, m.number);
    	return e.getMask().getBit(e.transformCache, mi, x, y);
	}
    public static boolean position_meeting(double x, double y, Entity e) {
    	e.validateTransform();
    	return positionMeetingImpl(x, y, e);
    }
    
    public static void position_destroy(double x, double y)
    {
    	Room.CURRENT.destroyAt(x, y);
    }
    
    /**Skips transformUpdates and self check*/
    protected static boolean collisionRectangleImpl(double x0, double y0, double x1, double y1, Entity e, boolean precise) {
    	if(e.transformCache.lx1 <= e.transformCache.lx0)
    		return false;
    	if(e.transformCache.ly1 <= e.transformCache.ly0)
    		return false;
    	
    	if(e.transformCache.x1 >= x0 && e.transformCache.y1 >= y0 && e.transformCache.x0 < x1 && e.transformCache.y0 < y1)
    	{
    		//Sic: Game maker uses the AABB for objects with not-transparent masks
    		if(precise && e.getMask().precise)
    		{
    			int ix0 = floor(Math.max(x0, e.transformCache.x0));
    	    	int ix1 = floor(Math.min(x1, e.transformCache.x1));
    	    	
    	    	int iy0 = floor(Math.max(y0, e.transformCache.y0));
    	    	int iy1 = floor(Math.min(y1, e.transformCache.y1));

    	    	Sprite m = e.getMask();
    	    	int mi = Modulo.mod((int)e.image_index, m.number);
    			for(int y = iy0; y <= iy1; y++)
    			{
    				for(int x = ix0; x <= ix1; x++)
    				{
    					if(m.getBit(e.transformCache, mi, x, y))
    						return true;
    				}
    			}
    			return false;
    		}
    		else
    			return true;
    	}
    	else
    		return false;
	}
    
    public<T extends Entity> T collision_rectangle(double x0, double y0, double x1, double y1, Class<T> obj, boolean precise, boolean notme)
	{
    	T r = null;
    	instances(meetingCache, obj);
    	for(int i = 0; i<meetingCache.size(); i++)
    	{
    		Entity e = meetingCache.get(i);
			if(e.getMask() == null)
					continue;
    		if(!notme || e != this)
    		{
	    		e.validateTransform();
	    		if(collisionRectangleImpl(x0, y0, x1, y1, e, precise))
	    		{
	    			r = obj.cast(e);
	    			break;
	    		}
    		}
    	}
    	meetingCache.clear();
		return r;
	}
    
    public void move_towards_point(double x, double y, double speed)
    {
    	double dx = x-this.x;
    	double dy = y-this.y;
    	double len = Trig.dist(dx, dy);
    	if(len < speed)
    	{
    		hspeed = 0;
    		vspeed = 0;
    		this.x = x;
    		this.y = y;
    	}
    	else
    	{
    		hspeed = dx*speed/len;
    		vspeed = dy*speed/len;
    	}
    }
    
    public void mp_potential_step(double x, double y, double stepsize, boolean includeNonSolid)
    {
    	final int stepsAhead = 3;
    	final double rotstep = 10;
    	double maxrot = 10;

    	double tdist = Trig.dist(this.x, this.y, x, y);
    	//Snap if already on the goal
    	if(tdist < 1)
    	{
    		this.x = x;
    		this.y = y;
    		hspeed = 0;
    		vspeed = 0;
    		return;
    	}
    	
    	double tdir = Trig.dir(this.x, this.y, x, y);
    	double sdist = getSpeed();
    	//If speed was zero, then assume we can go straight towards the goal.
    	double sdir = (sdist==0) ? tdir : Trig.dir(0, 0, hspeed, vspeed);
    	
    	double tsDiff = Modulo.mod(tdir-sdir, -180, 180);
    	//if target distance and direction are within range, then snap to target
    	if(tdist < stepsize && Math.abs(tsDiff) <= maxrot)
    	{
    		this.x = x;
    		this.y = y;
    		hspeed = 0;
    		vspeed = 0;
    		return;
    	}
    	//Compute the optimal amount of turning in order to go towards the goal
    	tsDiff = Vec1.min(tsDiff, maxrot);
    	
    	double dx =  stepsize*Trig.cos(sdir+tsDiff);
    	double dy = -stepsize*Trig.sin(sdir+tsDiff);
		int win = 2;//solution: 0=fail 1=partial 2=optimal
		for(int j = 1; j<=stepsAhead; j++)
		{
			if (!place_free(this.x + dx*j, this.y + dy*j, Object.class, !includeNonSolid))
			{
				if(j == 1)//If fail on the first step, then this direction is a failure
    			{
					win = 0;
    				break;
    			}
    			//otherwise partial solution
				win = 1;
			}
		}
    	//If the optimal works, then go for it maybe
		if (win > 0)
		{
			hspeed = dx;
			vspeed = dy;
			if(win == 2)//Straight for the goal is the optimal solution
				return;
		}
		if(sdist == 0)
			maxrot = 180;
		//We need to turn from current direction in order to avoid obstacles.
		//Decide which direction to prefer if we have a choice
		byte bias = (byte)(tsDiff<0 ? -1 : 1);
		for(double rot = rotstep; rot <= maxrot; rot+=rotstep)
		{
			byte bd = bias;
			for(int i = 0; i<2; i++)
			{
		    	dx =  stepsize*Trig.cos(sdir+bd*rot);
		    	dy = -stepsize*Trig.sin(sdir+bd*rot);
		    	int ok = 2;//assume optimal solution
		    	for(int j = 1; j<=stepsAhead; j++)
		    	{
		    		if(	!place_free(this.x + dx*j, this.y + dy*j, Object.class, !includeNonSolid))
		    		{
		    			if(j == 1)//If fail on the first step, then this direction is a failure
		    			{
		    				ok = 0;
		    				break;
		    			}
		    			//otherwise partial solution
		    			ok = 1;
		    		}
		    	}
		    	if(ok > 0)//If at least partial solution, then we can use this
				{
					hspeed = dx;
					vspeed = dy;
					if (ok == 2)//Search is over, we have a good solution
						return;
					win = 1;//Partial solution
				}
				bd = (byte) -bd;
			}
    	}
		if(win == 0)
		{
	    	//Failure, full stop
			hspeed = 0;
			vspeed = 0;
		}
    }
    
    public double distance_to_point(double x, double y)
    {
    	ObjectTransform tf = validateTransform();
    	double nx = Clamp.clamp(x, tf.x0, tf.x1);
    	double ny = Clamp.clamp(y, tf.y0, tf.y1);
    	return point_distance(nx, ny, x, y);
    }
    
    public double distance_to_object(Entity other)
    {
    	ObjectTransform a = validateTransform();
    	ObjectTransform b = other.validateTransform();
    	
    	double dx, dy;
    	if(b.x0 > a.x1)
    		dx = b.x0-a.x1;
    	else if(a.x0 > b.x1)
    		dx = a.x0-b.x1;
    	else
    		dx = 0;
    	if(b.y0 > a.y1)
    		dy = b.y0-a.y1;
    	else if(a.y0 > b.y1)
    		dy = a.y0-b.y1;
    	else
    		dy = 0;
    	return Trig.dist(dx, dy);
    }
	
	protected static final double[] mbs_projx = new double[8];
	protected static final double[] mbs_projy = new double[8];
	static 
	{
		mbs_projx[0] = 1;			mbs_projy[0] = 0;
		mbs_projx[1] = 0.7071;	mbs_projy[1] = -0.7071;
		mbs_projx[2] = 0;			mbs_projy[2] = -1;
		mbs_projx[3] = -0.7071;	mbs_projy[3] = -0.7071;
		mbs_projx[4] = -1;		mbs_projy[4] = 0;
		mbs_projx[5] = -0.7071; 	mbs_projy[5] = 0.7071;
		mbs_projx[6] = 0;			mbs_projy[6] = 1;
		mbs_projx[7] = 0.7071; 	mbs_projy[7] =  0.7071;
	}

    public void move_bounce_solid(boolean adv)
    {
    	move_bounce(adv, true);
    }
    
    public void move_bounce_all(boolean adv)
    {
    	move_bounce(adv, false);
    }
    
    public void move_bounce(boolean adv, boolean solid)
    {
    	if(adv)
    	{
    		if(!place_free(x+hspeed, y+vspeed, Object.class, solid))//Only bounce if actually heading into something
    		{
    		    //Project along various axes until free position
    			double sp = Trig.dist(hspeed, vspeed);
				for(int i = 0; i<sp+1; i++)
				{
					int free = 0;
					for(int j = 0; j < 8; j++)
					{
						if(place_free(x+hspeed+mbs_projx[j]*i, y+vspeed+mbs_projy[j]*i, Object.class, solid))
							free |= 1<<j;
					}
					
					if(free != 0)
					{
						//Search is over
						double ax = 0, ay = 0;
						int j = 0;
						while(free != 0)
						{
							if((free&1) != 0)
							{
								ax += mbs_projx[j];
								ay += mbs_projy[j];
							}
							free = free >>> 1;
							j++;
						}
						if(ax != 0 || ay != 0)
						{
							double ad = Trig.dist(ax, ay);
							ax /= ad;
							ay /= ad;
							double speedDotA = ax * hspeed + ay * vspeed;
							hspeed -= 2*speedDotA*ax;
							vspeed -= 2*speedDotA*ay;
						}
						return;
					}
				}
				//No solution found, bounce in both directions
				hspeed = -hspeed;
				vspeed = -vspeed;    			
    		}
    	}
    	else
    	{
    		if(!place_free(x+hspeed, y+vspeed, Object.class, solid))//Only bounce if actually heading into something
			{
    			byte sx = (byte) Math.signum(hspeed);
    			byte sy = (byte) Math.signum(vspeed);

    			//Project on x and y axes until the first free position
    			for(int i = 0; i<Math.max(Math.abs(hspeed), Math.abs(vspeed))+1; i++)
    			{
    				byte free = 0;
    				if(sx != 0 && place_free(x+hspeed-i*sx, y+vspeed, Object.class, solid))
    					free |= 0b10;
    				if(sy != 0 && place_free(x+hspeed, y+vspeed-i*sy, Object.class, solid))
    					free |= 0b01;    				
    				
    				if(free != 0)
    				{
    					//Search is over
    					if((free&0b10) != 0)
    						hspeed = -hspeed;
    					if((free&0b01) != 0)
    						vspeed = -vspeed;
    					return;
    				}
    			}
    			//No solution found, bounce in both directions
    			hspeed = -hspeed;
    			vspeed = -vspeed;
			}
    	}
    }
    
    public void move_contact_solid(double dir, double maxDist) {
    	move_contact(dir, maxDist, true);
    }
    public void move_contact_all(double dir, double maxDist) {
    	move_contact(dir, maxDist, false);
    }
    public void move_contact(double dir, double maxDist, boolean solid)
    {
    	double h = Trig.cos(dir);
    	double v = -Trig.sin(dir);
    	while(maxDist > 0.5)
    	{
    		if(!place_free(x+h, y+v, Object.class, solid))
    			return;
			x += h;
			y += v;
    		maxDist--;
    	}
    }
    
    protected static final double invsq2 = 1.0/Math.sqrt(2.0);
	
	protected static final double[] mv_dx = {
			-invsq2,	//SW
			0, 			//S
			invsq2, 	//SE 
			-invsq2, 	//W
			0, 			//C
			invsq2,  	//E 
			-invsq2, 	//NW
			0, 			//N
			invsq2  	//NE
	};
	protected static final double[] mv_dy = {
			invsq2, 	//SW
			invsq2,		//S
			invsq2,		//SE 
			0, 			//W
			0, 			//C 
			0, 			//E
			-invsq2, 	//NW
			-invsq2, 	//N
			-invsq2  	//NE
	};

	public final static short mv_sw = 0b100000000;
	public final static short mv_s  = 0b010000000;
	public final static short mv_se = 0b001000000;
	public final static short mv_w  = 0b000100000;
	public final static short mv_c  = 0b000010000;
	public final static short mv_e  = 0b000001000;
	public final static short mv_nw = 0b000000100;
	public final static short mv_n  = 0b000000010;
	public final static short mv_ne = 0b000000001;
    
    public void move(int dirMask, double speed)
    {
    	assert dirMask < 0b1000000000 : "dirMask out of range: "+Integer.toBinaryString(dirMask);
    	int bits = Integer.bitCount(dirMask);
    	int n = (int) Math.floor(Math.random()*bits);
    	int i = 0;
    	int j = 0;
    	while(dirMask != 0)
    	{
    		if((dirMask&1) != 0)
    		{
    			if(i==n)
    			{
    				hspeed = speed*mv_dx[8-j];
    				vspeed = speed*mv_dy[8-j];
    			}
    			i++;
    		}
    		dirMask >>>= 1;
				j++;
    	}
    	
    }
    public void motion_set(double dir, double sp)
    {
    	hspeed = lengthdir_x(sp, dir);
    	vspeed = lengthdir_y(sp, dir);
    }
    public void motion_add(double dir, double sp)
    {
    	hspeed += lengthdir_x(sp, dir);
    	vspeed += lengthdir_y(sp, dir);
    }
    public void setDirection(double newDirection)
    {
    	double sp = getSpeed();
    	hspeed = lengthdir_x(sp, newDirection);
    	vspeed = lengthdir_y(sp, newDirection);
    }
    public void setSpeed(double newSpeed)
    {
    	double speed = getSpeed();
    	if(speed == 0)
    		hspeed = newSpeed;
    	else
    	{
    		hspeed *= newSpeed/speed;
    		vspeed *= newSpeed/speed;
    	}
    }
    public void setDirSp(double newDirection, double newSpeed)
    {
    	hspeed = lengthdir_x(newSpeed, newDirection);
    	vspeed = lengthdir_y(newSpeed, newDirection);
    }
    
    public double getDirection()
    {
    	if(path_hspeed != 0 || path_vspeed != 0)
    		return Trig.dir(path_hspeed, path_vspeed);
    	return Modulo.mod(Trig.dir(hspeed, vspeed), 360.0);
    }
    public double getSpeed()
    {
    	return Trig.dist(hspeed, vspeed);
    }
    
    public double accelerate(double acc)
    {
    	double spd = getSpeed();
    	if(spd < 0.0000000001)
    	{
    		hspeed = Math.max(0, acc);
    		vspeed = 0;
    		return hspeed;
    	}
    	else
    	{
    		double newSpeed = Math.max(0, spd+acc);
	    	double mul = newSpeed/spd;
	    	hspeed *= mul;
	    	vspeed *= mul;
	    	return newSpeed;
    	}
    }
    
    /**Causes the cached ObjectTransform to update if necessary.
     * If it updates, the Quadtree position is also updated.*/
    public ObjectTransform validateTransform()
    {
    	if(!transformCache.matches(this) && !isDisposed())
		{
			transformCache.update(this);
		}
    	return transformCache;
    }
    
    /**bbox_left* Left side of the bounding box of the instance in the room, as defined by its image (taking scaling into account).*/
    public double bbox_left()
    {
    	validateTransform();
    	return transformCache.x0;
    }
    /**bbox_top* Top side of the bounding box of the instance in the room.*/
	public double bbox_top()
    {
    	validateTransform();
    	return transformCache.y0;
    }

    /**bbox_right* Right side of the bounding box of the instance in the room, minus 1.*/
	public double bbox_right()
    {
    	validateTransform();
    	return transformCache.x1-1;
    }
    /**bbox_bottom* Bottom side of the bounding box of the instance in the room, minus 1.*/
	public double bbox_bottom()
    {
    	validateTransform();
    	return transformCache.y1-1;
    }

	//dummy implementations mainly to make code generation easier
	protected void keyEvent(KeyEventType type, VK key) {}
	protected void mouseEvent(MouseEventType type, MB btn) {}
	
	protected void handleCollisions() {}

	public void endCollisions() {
		if(collisionSolid == 2) {
			x += hspeed;
			y += vspeed;
		}
		if(collisionSolid > 0 && !place_free(x, y)) {
			x = xprevious;
			y = yprevious;
		}
		collisionSolid = 0;
		collisionCache.clear();
	}

	public void activateNotify(boolean b)
	{
		this.active = b;
	}
	
	public boolean isActive()
	{
		return active;
	}

	public void addNotify(Room room)
	{
		rooms.add(room);
	}

	public void removeNotify(Room room)
	{
		rooms.remove(room);
	}
}
