package owg.lgm.gl;

import java.nio.FloatBuffer;
import java.nio.ShortBuffer;

import com.jogamp.opengl.GL2ES1;
import com.jogamp.opengl.util.GLBuffers;

import owg.lgm.gl.DrawCommand.DrawElements;
import owg.lgm.util.AutoDestructibleImpl;

public class TextImpl extends AutoDestructibleImpl implements Text
{
	public final BMFontRenderer font;
	
	protected int capacity;
	public FloatBuffer verts;
	protected ShortBuffer inds;
	protected transient ShortBuffer[] pageInds;
	
	protected boolean buffersChanged;
	protected final VertexBufferObjectF vbo;
	protected final IndexBufferObject ibo;
	
	protected final Mesh[] meshes;
	protected int meshCount;
	
	public TextImpl(GL2ES1 gl, BMFontRenderer font, int initialCapacity, BufferUsage usage)
	{
		assert font != null;
		this.font = font;
		allocate(initialCapacity);
		
		vbo = new VertexBufferObjectF(gl, verts, usage);
		autoDestroy(vbo);
		ibo = new IndexBufferObject(gl, inds, usage);
		autoDestroy(ibo);
		buffersChanged = false;

		meshes = new Mesh[font.pages.length];
		meshCount = 0;
		
		initMeshes();
	}
	
	protected void initMeshes()
	{
		AttribArray positionAttrib = new AttribArray(vbo, VertexAttribute.POSITION, 2, DataType.FLOAT, 4*4, 0);
		AttribArray texCoordAttrib = new AttribArray(vbo, VertexAttribute.TEXTURE_COORDINATE_0, 2, DataType.FLOAT, 4*4, 2*4);
		for(int i = 0; i<font.pages.length; i++)
		{
			meshes[i] = new Mesh(new DrawCommand.DrawElements(ibo, PrimitiveType.TRIANGLES, 0, DataType.UNSIGNED_SHORT, i*capacity*6*2),
					positionAttrib, texCoordAttrib);
			autoDestroy(meshes[i]);
		}
	}

	protected void allocate(int capacity)
	{
		this.capacity = capacity;
		verts = GLBuffers.newDirectFloatBuffer(capacity*4*4);
		inds = GLBuffers.newDirectShortBuffer(font.pages.length*capacity*6);
		inds.clear();
		pageInds = null;
	}

	@Override
	public void prepare(int length)
	{
		if(capacity < length) {
			allocate(Math.max(capacity*2, length));
			buffersChanged = true;
		} else {
			verts.clear();
		}
		
		if(pageInds == null)
		{
			pageInds = new ShortBuffer[font.pages.length];
			for(int i = 0; i<font.pages.length; i++) {
				inds.position(i*capacity*6);
				inds.limit((i+1)*capacity*6);
				pageInds[i] = inds.slice();
			}
		}
		else for(int i = 0; i<font.pages.length; i++)
			pageInds[i].clear();
	}
	
	@Override
	public void putChar(CharSequence str, int index, int pageIndex, float x0, float y0, float tx0, float ty0, float x1, float y1, float tx1, float ty1)
	{
		int k = verts.position()/4;
		
		verts.put(x0).put(y0).put(tx0).put(ty0);
		verts.put(x1).put(y0).put(tx1).put(ty0);
		verts.put(x0).put(y1).put(tx0).put(ty1);
		verts.put(x1).put(y1).put(tx1).put(ty1);
		
		ShortBuffer ibuf = pageInds[pageIndex];
		ibuf.put((short)(k+1)); ibuf.put((short)(k  )); ibuf.put((short)(k+2));
		ibuf.put((short)(k+1)); ibuf.put((short)(k+2)); ibuf.put((short)(k+3));
	}

	@Override
	public void finish()
	{
		int vCount = verts.position();
		if(vCount == 0)
		{
			meshCount = 0;
			return;
		}

		verts.clear();
		
		if(buffersChanged) {
			vbo.setData(verts);
			ibo.setData(inds);
			buffersChanged = false;
		} else {
			vbo.update(0, vCount);
			ibo.update();
		}
		
		for(int i = 0; i<font.pages.length; i++)
		{
			pageInds[i].flip();
			DrawElements draw = (DrawElements)meshes[i].drawCommand;
			draw.count = pageInds[i].remaining();
			draw.byteOffset = i*capacity*6*2;
		}
		meshCount = font.pages.length;
	}

	@Override
	public void render(GL2ES1 gl)
	{
		gl.glEnable(GL2ES1.GL_TEXTURE_2D);
    	for(int i = 0; i<meshCount; i++)
		{
			gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, font.pages[i].getHandle());
			meshes[i].render(gl);
		}
	}
}
