package owg.lgm.midi;

import owg.lgm.ResourceException;
import owg.lgm.plaf.Sound;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.LinkedBlockingQueue;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Sequence;
import javax.sound.midi.Soundbank;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;

import com.oddwarg.gervill.EnhancedSoftSynthesizer;

import static owg.competent.CompetentAudio.CA;

import owg.competent.CAClip;
import owg.competent.CompetentAudio;
import owg.competent.SampleFormat;
import owg.competent.SamplerState;
import owg.util.exceptionless.Join;
import owg.util.rethrow.Rethrow;

public class MidiThread extends Thread
{
	LinkedBlockingQueue<Runnable> evt = new LinkedBlockingQueue<>();

	File baseDir;
	Soundbank bank;
	AudioFormat format;
	EnhancedSoftSynthesizer synth;
	AudioInputStream stream;
	
	boolean paused = false;
	boolean dead = false;
	
	public String clientPlaying = null;
	Sequence sequence = null;
	long curtime;

	public final long samp;
	
	int buffersize = 2048; //frames
	ByteBuffer buf;
	CAClip clip;

	public MidiThread(File baseDir, File sf2)
	{
		super("midi-thread");
		setDaemon(true);
		
		this.baseDir = baseDir;
		
		try {
			bank = MidiSystem.getSoundbank(sf2);
			format = new AudioFormat(44100, 16, 2, true, false);
		} catch (IOException | InvalidMidiDataException err) {
			throw new ResourceException("Failed to load soundfont: "+sf2, err);
		}
		
		samp = CA.newSampler(Sound.sink);
		CompetentAudio.uninitializeListeners().add((ca) -> {
			dead = true;
			interrupt();
			Join.join(this);
		});
		
		// create buffer
		buf = ByteBuffer.allocateDirect(buffersize*4);
		buf.order(ByteOrder.LITTLE_ENDIAN);
		this.clip = new CAClip(2, 44100, SampleFormat.INT16, buf);
	}
	
	@Override
	public void run()
	{
		while(!dead) 
		{
			Runnable e; 
			while((e = evt.poll()) != null)
				e.run();
			
			byte[] tmp = new byte[buf.capacity()];
			if(!paused && synth != null)
			{
				if(sequence != null && synth.getMicrosecondPosition() >= curtime - 1000000L)
					curtime = synth.send(sequence, curtime);
				
				try {
					while(buf.hasRemaining())
					{
						int r = stream.read(tmp);
						if(r <= 0) {
							paused = true;
							break;
						}
						buf.put(tmp, 0, r);
					}
					buf.flip();
				
					CA.enqueueClip(samp, true, true, clip);
					while(CA.getClipQueueSize(samp) > 4)
					{
						if(CA.getState(samp) != SamplerState.PLAYING)
							CA.play(samp);
						try {
							Thread.sleep(33);
						} catch (InterruptedException e1) {
							return; //dead
						}
					}
				} catch (IOException err) {
					err.printStackTrace(); //unreachable
				}
			}
			else
			{
				try {
					evt.take().run();
				} catch (InterruptedException e1) {
					return; //dead
				}
			}
		}
	}

	public void midiPlay(String fname, boolean loop)
	{
		clientPlaying = fname;
		File f = new File(baseDir, fname);
		Sequence seq;
		try(InputStream fin = new FileInputStream(f)) {
			seq = MidiSystem.getSequence(fin);
		} catch(InvalidMidiDataException | IOException err) {
			throw new ResourceException("Failed to read midi file: "+fname);
		}
		
		evt.offer(() -> {
			try {
				if(synth != null)
				{
					stream.close();
					synth.close();
				}
				synth = new EnhancedSoftSynthesizer(bank);
				stream = synth.openStream(format, null);
			} catch (MidiUnavailableException | IOException err) {
				throw Rethrow.rethrow(err);
			}
			curtime = synth.send(seq, synth.getMicrosecondPosition());
			if(loop)
				sequence = seq;
			else
				sequence = null;
		});
	}
	
	public void midiStop()
	{
		clientPlaying = null;
		evt.offer(() -> {
			sequence = null;
			if(synth != null)
				synth.stop();
		});
	}
	
	public void midiPause()
	{
		evt.offer(() -> {
			paused = true;
		});
	}
	
	public void midiResume()
	{
		evt.offer(() -> {
			paused = false;
		});
	}
}
