package owg.lgm.plaf;

import owg.lgm.ResourceException;
import owg.lgm.util.FILE;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.util.HashMap;

import owg.competent.CAClip;
import owg.competent.CALoader;
import owg.competent.CompetentAudio;
import owg.competent.SamplerState;
import owg.competent.SoundModel;
import owg.competent.soundfile.CAClipLoader;
import owg.util.observables.DisposableImpl;

import static owg.competent.CompetentAudio.CA;
import static owg.lgm.plaf.Engine.EN;

public class Sound extends DisposableImpl
{
	protected static final HashMap<String, Sound> registry = new HashMap<>();
	
	public static final long sink, soundmixer;
	
	static
	{
		int bsz = 800;
		String prop = System.getProperty("ca.buffersize");
		if(prop != null)
		{
			try {
				bsz = Integer.parseInt(prop);
				System.out.println("CA buffer size adjusted to "+bsz);
			} catch(NumberFormatException err) {
				err.printStackTrace();
			}
		}
		
		CALoader.initialize();
		sink = CA.newSink(SoundModel.STEREO, bsz);
		soundmixer = CA.newMixer(sink);
		
		CA.openSink(sink, CompetentAudio.DEVICE_INDEX_AUTO, 44100);
	}
	
	public static void load(FILE current, int polyphony) throws IOException
	{
		CAClip clip = CAClipLoader.read(((JAVAFILE)current).file);
		if(clip != null)
		{
			Sound sound = new Sound(current.getName(), clip, polyphony);
			if(registry.put(sound.name, sound) != null)
				System.err.println("Sound name collision: "+sound.name);
			sound.addDisposeTask(() -> registry.remove(sound.name));
		}
	}
	
	public static Sound get(String name)
	{
		Sound r = registry.get(name);
		if(r == null)
			throw new ResourceException("No such Sound: "+name);
		return r;
	}
	public static void globalVolume(float value)
	{
		CA.setGain(soundmixer, 1.0/EN.animator.getTargetFPS(), value);
	}
	
	/**Convert from bad log scale to linear scale*/
	public static float linearize(double value)
	{
		return (float)(Math.pow(10, (value*2))-1)/99.0f;
	}
	
	public static void stopAll()
	{
		for(Sound s : registry.values())
			s.stop();
	}
	
	public final String name;
	public final CAClip clip;
	
	protected long[] samplers;
	protected int ptr = 0;
	
	
	protected float volume = 1.0f, pan = 0.0f;
	protected double speed = 1.0f;
	
	protected FloatBuffer channelMatrix;
	
	public Sound(String fname, CAClip clip, int polyphony)
	{	
		int end = fname.length();
		int sep = fname.lastIndexOf('.', end-2);
		while(sep != -1)
		{
			try {
				if(fname.charAt(sep+1) == 'v')
					volume = linearize(Integer.parseInt(fname.substring(sep+2, end))/100.0f);
				else if(fname.charAt(sep+1) == 'g')
					volume = Integer.parseInt(fname.substring(sep+2, end))/100.0f;
				else if(fname.charAt(sep+1) == 'p')
					pan = Integer.parseInt(fname.substring(sep+2, end))/100.0f;
				else if(fname.charAt(sep+1) == 'b')
					polyphony = Integer.parseInt(fname.substring(sep+2, end));
			} catch(NumberFormatException e) {
				//Don't care
			}
			end = sep;
			sep = fname.lastIndexOf('.', sep-1);
		}
		
		samplers = new long[polyphony];
		
		this.name = fname.substring(0, end);
		this.clip = clip;
		
		channelMatrix = ByteBuffer.allocateDirect(4*2*clip.getChannelCount()).order(ByteOrder.nativeOrder()).asFloatBuffer();
		fillChannelMatrix();
	}
	
	@Override
	public void dispose()
	{
		super.dispose();
		for(int i = 0; i<samplers.length; i++)
		{
			if(samplers[i] == 0)
				break;
			CA.deleteMachine(samplers[i]);
			samplers[i] = 0;
		}
	}
	
	protected void fillChannelMatrix()
	{
		channelMatrix.put(0, 						  Math.min(1.0f, 1.0f-pan)*volume);
		channelMatrix.put(2*clip.getChannelCount()-1, Math.min(1.0f, 1.0f+pan)*volume);
	}
	
	protected void selectSampler()
	{
		for(int i = 0; i<samplers.length; i++)
		{
			if(samplers[ptr] == 0) {
				samplers[ptr] = CA.newSampler(soundmixer);
				CA.enqueueClip(samplers[ptr], false, false, clip);
				CA.setChannelMatrix(samplers[ptr], 0, clip.getChannelCount(), 2, channelMatrix);
				CA.setFrequency(samplers[ptr], 0, speed);
				return;
			} else if(CA.getState(samplers[ptr]) != SamplerState.PLAYING)
				return;
			else
				ptr = (ptr+1)%samplers.length;
		}
		//No free sampler, use the next
		ptr = (ptr+1)%samplers.length;
	}

	public void play()
	{
		selectSampler();
		CA.setLooping(samplers[ptr], false);
		CA.play(samplers[ptr]);
	}
	
	public void loop()
	{
		selectSampler();
		CA.setLooping(samplers[ptr], true);
		CA.play(samplers[ptr]);
	}
	
	public void stop()
	{
		for(int i = 0; i<samplers.length && samplers[i] != 0; i++)
			CA.stop(samplers[i]);
	}
	
	public boolean isplaying()
	{
		for(int i = 0; i<samplers.length && samplers[i] != 0; i++) {
			if(CA.getState(samplers[i]) == SamplerState.PLAYING)
				return true;
		}
		return false;
	}
	
	public void volume(float value)
	{
		volume = value;
		fillChannelMatrix();
		for(int i = 0; i<samplers.length && samplers[i] != 0; i++)
			CA.setChannelMatrix(samplers[ptr], 1.0/EN.animator.getTargetFPS(), clip.getChannelCount(), 2, channelMatrix);
	}
	
	public void fade(float value, double millis)
	{
		volume = value;
		fillChannelMatrix();
		for(int i = 0; i<samplers.length && samplers[i] != 0; i++)
			CA.setChannelMatrix(samplers[ptr], millis/1000.0, clip.getChannelCount(), 2, channelMatrix);
	}
	
	public void pan(float value)
	{
		pan = value;
		fillChannelMatrix();
		for(int i = 0; i<samplers.length && samplers[i] != 0; i++)
			CA.setChannelMatrix(samplers[ptr], 1.0/EN.animator.getTargetFPS(), clip.getChannelCount(), 2, channelMatrix);
	}
}
