package owg.lgm.room;

import java.io.BufferedReader;
import java.io.EOFException;
import owg.lgm.util.FILE;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.NoSuchElementException;

import owg.lgm.ResourceException;
import owg.lgm.plaf.Engine;
import owg.util.fileextensions.FileExtensions;

public class RoomData
{
	protected static final String UTF_8 = "UTF-8";
	protected static final String ISO_8859_1 = "ISO-8859-1";
	
	public static ArrayList<RoomData> list = new ArrayList<>();
	public static HashMap<String, RoomData> map = new HashMap<>();

	public static RoomData get(String room)
	{
		RoomData r = map.get(room);
		if(r == null)
			throw new NoSuchElementException(room);
		return r;
	}
	
	public static boolean readDir(FILE file) {
		//Order matters here, so use the toc.txt which was kindly provided by egm
		String fname = "toc.txt";
		FILE toc = file.getChild(fname);
		if(!toc.isFile())
			return false;
		
		try (BufferedReader br = new BufferedReader(new InputStreamReader(toc.read()))) {
			while((fname = br.readLine()) != null)
			{
				FILE df = file.getChild(fname);
				boolean dir = readDir(df);
				if(!dir)
					readRoom(file, fname);
			}
			return true;
		} catch (IOException e) {
			throw new ResourceException("Error reading rooms at: "+file+"/"+fname, e);
		}
	}
	protected static void readRoom(FILE dir, String f) throws IOException {
		//System.out.println("Loading Room: "+dir+"/"+f);
		FILE ey = dir.getChild(f+".ey");
		RoomData room = new RoomData();
		room.dir = dir;
		
		try(BufferedReader br = new BufferedReader(new InputStreamReader(ey.read())))
		{
			RoomInfo info = new RoomInfo();
			info.name = f;
			String line;
			while((line = br.readLine()) != null)
			{
				int sep = line.indexOf(':');
				if(sep != -1)
				{
					String key = line.substring(0, sep).trim();
					String val = line.substring(sep+1).trim();
					switch(key)
					{
						case "Data":
							val = FileExtensions.stripFileExtension(val);
							if(!val.equals(f))
								System.err.println("Declared name "+val+" does not match file name "+f);
							break;
						case "CAPTION":
							info.caption = val;
							break;
						case "WIDTH":
							info.width = Integer.parseInt(val);
							break;
						case "HEIGHT":
							info.height = Integer.parseInt(val);
							break;
						case "SPEED":
							info.fps = Integer.parseInt(val);
							break;
						case "PERSISTENT":
							info.persistent = Boolean.parseBoolean(val);
							break;
						case "BACKGROUND_COLOR":
							if(!val.startsWith("0x") && !val.startsWith("0X"))
								throw new IOException("Expected hex string starting with 0x, at: "+line);
							info.backgroundColor = Integer.parseUnsignedInt(val.substring(2), 16);
							break;
						case "DRAW_BACKGROUND_COLOR":
							info.drawBackgroundColor = Boolean.parseBoolean(val);
							break;
						case "VIEWS_ENABLED":
							info.viewsEnabled = Boolean.parseBoolean(val);
							break;
						case "SNAP_X":
							info.snapX = Integer.parseInt(val);
							break;
						case "SNAP_Y":
							info.snapY = Integer.parseInt(val);
							break;
						case "SHOW_GRID":
							info.showGrid = Boolean.parseBoolean(val);
							break;
						case "SCROLL_BAR_X":
							info.scrollBarX = Integer.parseInt(val);
							break;
						case "SCROLL_BAR_Y":
							info.scrollBarY = Integer.parseInt(val);
							break;
						default:
							break;
					}
				}
			}
			room.info = info;
		}

		FILE rmg = dir.getChild(f+".rmg");
		try (InputStream fis = rmg.read();)
		{
			ByteBuffer readBuf = ByteBuffer.allocate(8);
			readBuf.order(ByteOrder.LITTLE_ENDIAN);
			
			//Not actually used, but keep it for sanity
			room.creationCode = readString(fis, readBuf, UTF_8);
			
			int backgroundDefsSize = readInt4(fis, readBuf);
			room.backgrounds = new BackgroundInfo[backgroundDefsSize];
			for(int i = 0; i<room.backgrounds.length; i++)
			{
				BackgroundInfo background = new BackgroundInfo();
				background.visible = readBoolean4(fis, readBuf);
				background.foreground = readBoolean4(fis, readBuf);
				background.backgroundName = readString(fis, readBuf, ISO_8859_1);
				background.x = readInt4(fis, readBuf);
				background.y = readInt4(fis, readBuf);
				background.tileX = readBoolean4(fis, readBuf);
				background.tileY = readBoolean4(fis, readBuf);
				background.hspeed = readInt4(fis, readBuf);
				background.vspeed = readInt4(fis, readBuf);
				background.stretch = readBoolean4(fis, readBuf);
				room.backgrounds[i] = background;
			}
			
			int viewsSize = readInt4(fis, readBuf);
			
			room.views = new ViewInfo[viewsSize];
			for(int i = 0; i<room.views.length; i++)
			{
				ViewInfo view = new ViewInfo();
				view.visible = readBoolean4(fis, readBuf);
				view.xview = readInt4(fis, readBuf);
				view.yview = readInt4(fis, readBuf);
				view.wview = readInt4(fis, readBuf);
				view.hview = readInt4(fis, readBuf);
				view.xport = readInt4(fis, readBuf);
				view.yport = readInt4(fis, readBuf);
				view.wport = readInt4(fis, readBuf);
				view.hport = readInt4(fis, readBuf);
				
				view.hborder = readInt4(fis, readBuf);
				view.vborder = readInt4(fis, readBuf);
				view.hspeed = readInt4(fis, readBuf);
				view.vspeed = readInt4(fis, readBuf);
				view.objectName = readString(fis, readBuf, ISO_8859_1);
				room.views[i] = view;
			}
			
			
			int instancesSize = readInt4(fis, readBuf);
			room.instances = new InstanceInfo[instancesSize];
			for(int i = 0; i<room.instances.length; i++)
			{
				InstanceInfo instance = new InstanceInfo();
				instance.x = readInt4(fis, readBuf);
				instance.y = readInt4(fis, readBuf);
				instance.xscale = readDouble8(fis, readBuf);
				instance.yscale = readDouble8(fis, readBuf);
				instance.color = readInt4(fis, readBuf);
				instance.angle = readDouble8(fis, readBuf);
				instance.object = readString(fis, readBuf, ISO_8859_1);
				instance.id = readInt4(fis, readBuf);
				instance.creationCode = readString(fis, readBuf, UTF_8);
				instance.locked = readBoolean4(fis, readBuf);
				room.instances[i] = instance;
			}			
			
			int tilesSize = readInt4(fis, readBuf);
			room.tiles = new TileInfo[tilesSize];
			for(int i = 0; i<room.tiles.length; i++)
			{
				TileInfo tile = new TileInfo();
				tile.x = readInt4(fis, readBuf);
				tile.y = readInt4(fis, readBuf);
				tile.backgroundName = readString(fis, readBuf, ISO_8859_1);
				tile.backgroundX = readInt4(fis, readBuf);
				tile.backgroundY = readInt4(fis, readBuf);
				tile.backgroundW = readInt4(fis, readBuf);
				tile.backgroundH = readInt4(fis, readBuf);
				tile.depth = readInt4(fis, readBuf);
				tile.id = readInt4(fis, readBuf);
				tile.locked = readBoolean4(fis, readBuf);
				room.tiles[i] = tile;
			}
		}
		
		room.index = list.size();
		list.add(room);
		map.put(room.info.name, room);
	}

	protected static String readString(InputStream fis, ByteBuffer readBuf, String charSet) throws IOException {
		int strLen = readInt4(fis, readBuf);
		byte[] data = new byte[strLen];
		readFully(fis, data, 0, strLen);
		return new String(data, charSet);
	}
	protected static int readInt4(InputStream fis, ByteBuffer readBuf) throws IOException {
		readFully(fis, readBuf.array(), 0, 4);
		int result = readBuf.getInt();
		readBuf.rewind();
		return result;
	}
	protected static double readDouble8(InputStream fis, ByteBuffer readBuf) throws IOException {
		readFully(fis, readBuf.array(), 0, 8);
		double result = readBuf.getDouble();
		readBuf.rewind();
		return result;
	}
	protected static boolean readBoolean4(InputStream fis, ByteBuffer readBuf) throws IOException {
		readFully(fis, readBuf.array(), 0, 4);
		int result = readBuf.getInt();
		readBuf.rewind();
		return result != 0;
	}
	protected static void readFully(InputStream stream, byte[] bytes, int off, int len) throws IOException {
		int total = 0;
		while(total < len)
		{
			int read = stream.read(bytes, off+total, len-total);
			if(read == -1)
				throw new EOFException("Reached the end of the stream before "+len+" bytes could be read(only got "+total+")");
			total += read;
		}
	}

	public static void mangleCreationCodes()
	{
		for(int j = 0; j<list.size(); j++)
		{
			RoomData r = list.get(j);
			int ccCount = 0;
			for(int i = 0; i<r.instances.length; i++)
			{
				String creationCode = r.instances[i].creationCode;
				cc: if(!creationCode.isEmpty())
				{
					ccCount++;
					r.instances[i].creationCode = r.info.name+"#"+ccCount;
					Class<?> instanceClass = Engine.EN.gameConfig.objects.get(r.instances[i].object);
					if(instanceClass != null)
					{
						CCHandled a = instanceClass.getAnnotation(CCHandled.class);
						if(a != null)
						{
							for(String v : a.value())
								if(v.equals(r.instances[i].creationCode))
									break cc;
						}
						System.err.println("Warning: Unhandled creation code for "+r.instances[i].object+
								", ccId=\""+r.instances[i].creationCode+"\"");
						System.err.println(creationCode);
					}
				}
			}
				
		}
	}

	public void writeRoom() throws IOException {
		FILE ey = dir.getChild(info.name+".ey");
		String rmgName = info.name+".rmg";
		
		try(PrintWriter pw = new PrintWriter(ey.write()))
		{
			pw.println("Data:"+rmgName);
			pw.println("CAPTION:"+info.caption);
			pw.println("WIDTH:"+info.width);
			pw.println("HEIGHT:"+info.height);
			pw.println("SPEED:"+info.fps);
			pw.println("PERSISTENT:"+info.persistent);
			pw.println("BACKGROUND_COLOR:0x"+Integer.toHexString(info.backgroundColor));
			pw.println("DRAW_BACKGROUND_COLOR:"+info.drawBackgroundColor);
			pw.println("VIEWS_ENABLED:"+info.viewsEnabled);

			pw.println("SNAP_X:"+info.snapX);
			pw.println("SNAP_Y:"+info.snapY);
			pw.println("SHOW_GRID:"+info.showGrid);
			pw.println("SCROLL_BAR_X:"+info.scrollBarX);
			pw.println("SCROLL_BAR_Y:"+info.scrollBarY);
		}

		FILE rmg = dir.getChild(rmgName);
		try (OutputStream fos = rmg.write();)
		{
			ByteBuffer writeBuf = ByteBuffer.allocate(8);
			writeBuf.order(ByteOrder.LITTLE_ENDIAN);
			
			writeString(fos, writeBuf, creationCode, UTF_8);
			
			writeInt4(fos, writeBuf, backgrounds.length);
			for(int i = 0; i<backgrounds.length; i++)
			{
				BackgroundInfo background = backgrounds[i];
				writeBoolean4(fos, writeBuf, background.visible);
				writeBoolean4(fos, writeBuf, background.foreground);
				writeString(fos, writeBuf, background.backgroundName, ISO_8859_1);
				writeInt4(fos, writeBuf, background.x);
				writeInt4(fos, writeBuf, background.y);
				writeBoolean4(fos, writeBuf, background.tileX);
				writeBoolean4(fos, writeBuf, background.tileY);
				writeInt4(fos, writeBuf, background.hspeed);
				writeInt4(fos, writeBuf, background.vspeed);
				writeBoolean4(fos, writeBuf, background.stretch);
			}
			
			writeInt4(fos, writeBuf, views.length);
			for(int i = 0; i<views.length; i++)
			{
				ViewInfo view = views[i];
				writeBoolean4(fos, writeBuf, view.visible);
				writeInt4(fos, writeBuf, view.xview);
				writeInt4(fos, writeBuf, view.yview);
				writeInt4(fos, writeBuf, view.wview);
				writeInt4(fos, writeBuf, view.hview);
				writeInt4(fos, writeBuf, view.xport);
				writeInt4(fos, writeBuf, view.yport);
				writeInt4(fos, writeBuf, view.wport);
				writeInt4(fos, writeBuf, view.hport);
				
				writeInt4(fos, writeBuf, view.hborder);
				writeInt4(fos, writeBuf, view.vborder);
				writeInt4(fos, writeBuf, view.hspeed);
				writeInt4(fos, writeBuf, view.vspeed);
				writeString(fos, writeBuf, view.objectName, ISO_8859_1);
			}
			
			writeInt4(fos, writeBuf, instances.length);
			for(int i = 0; i<instances.length; i++)
			{
				InstanceInfo instance = instances[i];
				writeInt4(fos, writeBuf, instance.x);
				writeInt4(fos, writeBuf, instance.y);
				writeDouble8(fos, writeBuf, instance.xscale);
				writeDouble8(fos, writeBuf, instance.yscale);
				writeInt4(fos, writeBuf, instance.color);
				writeDouble8(fos, writeBuf, instance.angle);
				writeString(fos, writeBuf, instance.object, ISO_8859_1);
				writeInt4(fos, writeBuf, instance.id);
				writeString(fos, writeBuf, instance.creationCode, UTF_8);
				writeBoolean4(fos, writeBuf, instance.locked);
			}			

			writeInt4(fos, writeBuf, tiles.length);
			for(int i = 0; i<tiles.length; i++)
			{
				TileInfo tile = tiles[i];
				writeInt4(fos, writeBuf, tile.x);
				writeInt4(fos, writeBuf, tile.y);
				writeString(fos, writeBuf, tile.backgroundName, ISO_8859_1);
				writeInt4(fos, writeBuf, tile.backgroundX);
				writeInt4(fos, writeBuf, tile.backgroundY);
				writeInt4(fos, writeBuf, tile.backgroundW);
				writeInt4(fos, writeBuf, tile.backgroundH);
				writeInt4(fos, writeBuf, tile.depth);
				writeInt4(fos, writeBuf, tile.id);
				writeBoolean4(fos, writeBuf, tile.locked);
			}
		}
	}

	protected static void writeString(OutputStream fos, ByteBuffer writeBuf, String val, String charSet) throws IOException {
		byte[] data = val.getBytes(charSet);
		writeInt4(fos, writeBuf, data.length);
		fos.write(data);
	}
	protected static void writeInt4(OutputStream fos, ByteBuffer writeBuf, int val) throws IOException {
		writeBuf.putInt(val);
		fos.write(writeBuf.array(), 0, 4);
		writeBuf.rewind();
	}
	protected static void writeDouble8(OutputStream fos, ByteBuffer writeBuf, double val) throws IOException {
		writeBuf.putDouble(val);
		fos.write(writeBuf.array(), 0, 8);
		writeBuf.rewind();
	}
	protected static void writeBoolean4(OutputStream fos, ByteBuffer writeBuf, boolean val) throws IOException {
		writeBuf.putInt(val ? 1 : 0);
		fos.write(writeBuf.array(), 0, 4);
		writeBuf.rewind();
	}
	
	public static class RoomInfo {
		public String name;
		public String caption; 
		public int width;
		public int height;
		public int fps;
		public boolean persistent;
		public int backgroundColor;
		public boolean drawBackgroundColor;
		public boolean viewsEnabled;
		
		public int snapX;
		public int snapY;
		public boolean showGrid;
		public int scrollBarX;
		public int scrollBarY;
	}
	
	public static class BackgroundInfo {

		public boolean visible;
		public boolean foreground;
		protected String backgroundName;
		public int x;
		public int y;
		public int hspeed;
		public int vspeed;
		public boolean tileX;
		public boolean tileY;
		public boolean stretch;
		public double alpha = 1.0;
		
		public String getBackgroundName()
		{
			return backgroundName;
		}
		public void setBackgroundName(String backgroundName)
		{
			this.backgroundName = backgroundName;
		}
	}
	
	public static class ViewInfo {

		public boolean visible;
		public int xview;
		public int yview;
		public int wview;
		public int hview;
		public int xport;
		public int yport;
		public int wport;
		public int hport;
		
		public int hborder;
		public int vborder;
		public int hspeed;
		public int vspeed;

		protected String objectName;
		
		public String getObjectName()
		{
			return objectName;
		}
		public void setObjectName(String objectName)
		{
			this.objectName = objectName;
		}
	}
	
	public static class InstanceInfo {
		public int x;
		public int y;
		public double xscale;
		public double yscale;
		public int color;
		public double angle;
		public String object;
		public int id;
		public String creationCode;
		public boolean locked;
	}
	
	public static class TileInfo {
		public int x;
		public int y;
		public String backgroundName;
		public int backgroundX;
		public int backgroundY;
		public int backgroundW;
		public int backgroundH;
		public int depth;
		public int id;
		public boolean locked;
	}
	
	public FILE dir;
	public int index;
	
	public RoomInfo info;
	public String creationCode;
	public BackgroundInfo[] backgrounds;
	public ViewInfo[] views;
	public InstanceInfo[] instances;
	public TileInfo[] tiles;
	
	@Override
	public String toString()
	{
		return info.name;
	}
}
