package owg.competent;

import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;

/**<p>
 *A clip consists of a series of numbers representing the amplitudes of an audio signal at uniform intervals (PCM).
 *</p><p>
 *These numbers are called samples. The CASampleFormat enum describes the supported data types for samples.
 *For data types that take more than 1 byte (8 bits), there is no agreement over whether the most significant byte shall
 *be stored first (big-endian) or last (little-endian). For this reason, you must also specify the byte order.<ul>
 *<li>Almost all audio files use little-endian. If you got the data from a file then it is likely little-endian.</li>
 *<li>When generating sample data, it is best to use a native-endian {@link Buffer} (see {@link ByteOrder#nativeOrder()}).</li>
 *</ul>
 *</p><p>
 *The sample rate determines the interval between the samples, in seconds: interval = 1.0 / sample rate.
 *</p><p>
 *Multi-channel sample data has interleaved channels. For instance, the following series represents a stereo signal:
 *1st left sample, 1st right sample, 2nd left sample, 2nd right sample, 3rd left sample, 3rd right sample, ...
 *A block of one sample for each channel is called a sample frame.
 *</p><p>
 *If you need to, you can control how sample channels are mapped to output channels using mixing matrices.
 *</p><p>*/
public class CAClip 
{
	protected int channelCount;
	protected double sampleRate; 
	protected SampleFormat sampleFormat; 
	protected ByteOrder byteOrder;
	protected int frameCount;
	protected final Buffer sampleData;

	/**Creates a clip object using the byte order of the sampleData buffer, and a length inferred from its capacity.<ul>
	 * <li>sampleData must be a <b>direct buffer</b>, i.e. it must have been allocated using {@link ByteBuffer#allocateDirect(int)}.</li>
	 * <li>Enqueueing the clip to a sampler will <b>convert the sample data to the native byte order</b> unless the copyData mode is used:
	 * If sampleData does not already use {@link ByteOrder#nativeOrder()}, its contents may become jumbled.</li>
	 * <li>The application <b>must not modify the buffer</b> after enqueueing the clip to a sampler, unless the copyData mode is used.
	 * </ul>
	 * @param channelCount specifies the number of channels in the sample data.
	 * @param sampleRate specifies the sample rate for the clip at normal playback speed.
	 * @param sampleFormat specifies the data type for the samples in the sampleData array.
	 * @param sampleData is the array of samples. For multichannel audio, an interleaved format must be used.
	 * 
	 * @throws CAException If the sampleData buffer is not a direct buffer.
	 */
	public CAClip(int channelCount, double sampleRate, SampleFormat sampleFormat, ByteBuffer sampleData) throws CAException {
		this(channelCount, sampleRate, sampleFormat, sampleData.order(), sampleData.capacity()/(channelCount*sampleFormat.sizeInBytes), sampleData);
	}
	
	/**Creates a clip object using the byte order and format given by the sampleData buffer, and a length inferred from its capacity.<ul>
	 * <li>sampleData must be a <b>direct buffer</b>, i.e. it must have been allocated using {@link ByteBuffer#allocateDirect(int)}.</li>
	 * <li>Enqueueing the clip to a sampler will <b>convert the sample data to the native byte order</b> unless the copyData mode is used:
	 * If sampleData does not already use {@link ByteOrder#nativeOrder()}, its contents may become jumbled.</li>
	 * <li>The application <b>must not modify the buffer</b> after enqueueing the clip to a sampler, unless the copyData mode is used.
	 * </ul>
	 * @param channelCount specifies the number of channels in the sample data.
	 * @param sampleRate specifies the sample rate for the clip at normal playback speed.
	 * @param sampleData is the array of samples. For multichannel audio, an interleaved format must be used.
	 * 
	 * @throws CAException If the sampleData buffer is not a direct buffer.
	 */
	public CAClip(int channelCount, double sampleRate, ShortBuffer sampleData) throws CAException {
		this(channelCount, sampleRate, SampleFormat.INT16, sampleData.order(), sampleData.capacity()/channelCount, sampleData);
	}
	
	/**Creates a clip object using the byte order and format given by the sampleData buffer, and a length inferred from its capacity.<ul>
	 * <li>sampleData must be a <b>direct buffer</b>, i.e. it must have been allocated using {@link ByteBuffer#allocateDirect(int)}.</li>
	 * <li>Enqueueing the clip to a sampler will <b>convert the sample data to the native byte order</b> unless the copyData mode is used:
	 * If sampleData does not already use {@link ByteOrder#nativeOrder()}, its contents may become jumbled.</li>
	 * <li>The application <b>must not modify the buffer</b> after enqueueing the clip to a sampler, unless the copyData mode is used.
	 * </ul>
	 * @param channelCount specifies the number of channels in the sample data.
	 * @param sampleRate specifies the sample rate for the clip at normal playback speed.
	 * @param sampleData is the array of samples. For multichannel audio, an interleaved format must be used.
	 * 
	 * @throws CAException If the sampleData buffer is not a direct buffer.
	 */
	public CAClip(int channelCount, double sampleRate, FloatBuffer sampleData) throws CAException {
		this(channelCount, sampleRate, SampleFormat.FLOAT32, sampleData.order(), sampleData.capacity()/channelCount, sampleData);
	}
	/**Creates a clip object.<ul>
	 * <li>sampleData must be a <b>direct buffer</b>, i.e. it must have been allocated using {@link ByteBuffer#allocateDirect(int)}.</li>
	 * <li>Enqueueing the clip to a sampler will <b>convert the sample data to the native byte order</b> unless the copyData mode is used:
	 * If byteOrder is not already {@link ByteOrder#nativeOrder()}, its contents may become jumbled.</li>
	 * <li>The application <b>must not modify the buffer</b> after enqueueing the clip to a sampler, unless the copyData mode is used.
	 * </ul>
	 * @param channelCount specifies the number of channels in the sample data.
	 * @param sampleRate specifies the sample rate for the clip at normal playback speed.
	 * @param sampleFormat specifies the data type for the samples in the sampleData array.
	 * @param byteOrder specifies the current byte order of the data in the sampleData array.
	 * @param frameCount specifies the number of sample frames in the sampleData array, where a frame is a block of channelCount samples.
	 * @param sampleData is the array of samples. For multichannel audio, an interleaved format must be used.
	 * 
	 * @throws CAException If the sampleData buffer is not a direct buffer.
	 */
	public CAClip(int channelCount, double sampleRate, SampleFormat sampleFormat, ByteOrder byteOrder, int frameCount, Buffer sampleData) throws CAException {
		if(!sampleData.isDirect())
			throw new CAException(CAErrorCode.ERROR_INVALID, "sampleData.isDirect()=false");
		this.channelCount = channelCount;
		this.sampleRate = sampleRate;
		this.sampleFormat = sampleFormat;
		this.byteOrder = byteOrder;
		this.frameCount = frameCount;
		this.sampleData = sampleData;
	}
	
	public int getChannelCount()
	{
		return channelCount;
	}
	public double getSampleRate()
	{
		return sampleRate;
	}
	public SampleFormat getSampleFormat()
	{
		return sampleFormat;
	}
	/**Sets the current byte order to the given byte order. This does not modify the sample data.*/
	protected void setByteOrder(ByteOrder order)
	{
		this.byteOrder = order;
	}
	/**Returns the byte order that is used when interpreting this clip's sample data.*/
	public ByteOrder getByteOrder()
	{
		return byteOrder;
	}
	/**Returns the number of frames in the sample data. (A frame contains one sample for each channel.)*/
	public int getFrameCount()
	{
		return frameCount;
	}
	/**Gets the sample data buffer. This is a direct buffer.*/
	public Buffer getSampleData() 
	{
		return sampleData;
	}
}
