package owg.competent.soundfile;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.ShortBuffer;

import javazoom.jl.decoder.Bitstream;
import javazoom.jl.decoder.BitstreamException;
import javazoom.jl.decoder.Decoder;
import javazoom.jl.decoder.DecoderException;
import javazoom.jl.decoder.Header;
import javazoom.jl.decoder.SampleBuffer;
import owg.competent.SampleFormat;

public class MP3InputStream extends AudioDecoderStream
{
	protected final Bitstream bitstream;
	protected final Decoder decoder;
	
	ByteBuffer sampleBytes;
	ShortBuffer sampleShorts;
	
	protected int sampleRate;
	protected int channelCount;
	protected byte[] dummy;
	
	public MP3InputStream(InputStream in) throws IOException {
		super(in);
		
        bitstream = new Bitstream(in);
        decoder = new Decoder();
    	Header header;
		try
    	{
	        header = bitstream.readFrame();
	        if(header == null)
	        	throw new IOException("Failed to find mp3 header");
    	} catch(BitstreamException e) {
    		throw new IOException("Failed to read mp3 header", e);
    	}
        
    	SampleBuffer buffer;
		try {
    		buffer = (SampleBuffer)decoder.decodeFrame(header, bitstream);
    	} catch(DecoderException e) {
    		throw new IOException("Failed to decode mp3 frame", e);
    	}
        sampleBytes = ByteBuffer.allocate(decoder.getOutputBlockSize()*2);
        sampleBytes.order(ByteOrder.LITTLE_ENDIAN);
        sampleShorts = sampleBytes.asShortBuffer();
        sampleShorts.put(buffer.getBuffer(), 0, buffer.getBufferLength());
        sampleShorts.flip();
        sampleBytes.limit(sampleShorts.limit()*2);
        
        sampleRate = decoder.getOutputFrequency();
        channelCount = decoder.getOutputChannels();
        
        bitstream.closeFrame();
		dummy = new byte[1];
	}
	
	protected void rebuffer() throws IOException
	{
		try {
			Header header = bitstream.readFrame();
			if(header == null)
			{
				sampleShorts.rewind();
				sampleShorts.limit(0);
				sampleBytes.rewind();
				sampleBytes.limit(0);
				return;
			}
			SampleBuffer buffer = (SampleBuffer)decoder.decodeFrame(header, bitstream);
	        sampleShorts.put(buffer.getBuffer(), 0, buffer.getBufferLength());
	        sampleShorts.flip();
	        sampleBytes.rewind();
	        sampleBytes.limit(sampleShorts.limit()*2);
			
			bitstream.closeFrame();
		} catch(BitstreamException e) {
    		throw new IOException("Failed to read mp3 header", e);
		} catch(DecoderException e) {
    		throw new IOException("Failed to decode mp3 frame", e);
		}
	}

	@Override
	public int read(byte[] b, int off, int len) throws IOException {
		if(!sampleBytes.hasRemaining())
			return -1; //End of file
		
		int read = Math.min(len, sampleBytes.remaining());
		sampleBytes.get(b, off, read);
		
		if(!sampleBytes.hasRemaining())
			rebuffer();
		return read;
	}

	@Override
	public int read(byte[] b) throws IOException {
		return read(b, 0, b.length);
	}
	
	@Override
	public int read() throws IOException {
		int rd = read(dummy, 0, 1);
		if(rd == -1)
			return -1;
		return dummy[0];
	}
	
	@Override
	public int available() throws IOException {
		return sampleBytes.remaining();
	}
	@Override
	public boolean markSupported() {
		return false;
	}
	
	@Override
	public synchronized void mark(int readlimit) {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public synchronized void reset() throws IOException {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public long skip(long n) throws IOException {
		byte[] skip = new byte[(int) Math.min(n, 4096L)];
		return read(skip);
	}

	@Override
	public double getSampleRate() {
		return sampleRate;
	}

	@Override
	public int getChannelCount() {
		return channelCount;
	}

	@Override
	public SampleFormat getSampleFormat() {
		//mp3 is always 16 bit
		return SampleFormat.INT16;
	}

	@Override
	public boolean isLittleEndian() {
		return true;
	}
	
}